local AddonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensSpellFlashLibrary

function a.GetRegen()
    return select(2, GetPowerRegen())
end

function a.InSoloMode()
    return (not s.InRaidOrParty()) and (not a.GetConfig("solo_off"))
end

------------------------------------------------------------------------ Common
a.spells["Vanish"] = { ID = 1856 }

a.EquipmentSets = {}
a.EquipmentSets.T13 = {
    HeadSlot = { 78794, 77025, 78699 },
    ShoulderSlot = { 78833, 77027, 78738 },
    ChestSlot = { 78759, 77023, 78664 },
    HandsSlot = { 78774, 77024, 78679 },
    LegsSlot = { 78803, 77026, 78708 },
}

local function poisonNeeded(names, offhand)
    local min
    if s.InCombat() then
        min = 0
    else
        min = 5 * 60
    end
    if offhand then
        return not s.OffHandItemBuff(names, min)
    else
        return not s.MainHandItemBuff(names, min)
    end
end

a.spells["Instant Poison"] = {
    Type = "item",
    ID = 6947,
    FlashColor = "yellow",
    CheckFirst = function()
        return poisonNeeded(L["Instant Poison"], false)
    end
}

local instantOrWound = { L["Instant Poison"], L["Wound Poison"] }
a.spells["Instant or Wound"] = {
    Type = "item",
    ID = 10918,
    FlashID = { 6947, 10918 },
    FlashColor = "yellow",
    CheckFirst = function()
        return poisonNeeded(instantOrWound, false)
    end
}

a.spells["Deadly Poison"] = {
    Type = "item",
    ID = 2892,
    FlashColor = "yellow",
    CheckFirst = function()
        return poisonNeeded(L["Deadly Poison"], true)
    end
}

local deadlyOrWound = { L["Deadly Poison"], L["Wound Poison"] }
a.spells["Deadly or Wound"] = {
    Type = "item",
    ID = 2892,
    FlashID = { 2892, 10918 },
    FlashColor = "yellow",
    CheckFirst = function()
        return poisonNeeded(deadlyOrWound, true)
    end
}

a.spells["Slice and Dice"] = {
    ID = 5171,
    Buff = 5171,
    BuffUnit = "player",
    NoRangeCheck = 1,
}

a.spells["Slice and Dice unless Solo"] = {
    ID = 5171,
    Buff = 5171,
    BuffUnit = "player",
    NoRangeCheck = 1,
    CheckFirst = function()
        return not a.InSoloMode()
    end
}

a.spells["Recuperate"] = {
    ID = 73651,
    NoRangeCheck = 1,
    FlashColor = "yellow",
}

a.spells["Sinister Strike"] = {
    ID = 1752,
}

a.spells["Tricks of the Trade"] = {
    ID = 57934,
    FlashColor = "yellow",
    NoRangeCheck = 1,
    CheckFirst = function()
        return s.InRaidOrParty()
            and (not c.WearingSet(2, "T13")
                or not (c.HasBuff("Adrenaline Rush")
                    or s.Buff(c.BLOODLUST_BUFFS, "player")))
    end
}

a.spells["Kick"] = {
    ID = 1766,
    Interrupt = 1,
    FlashColor = "aqua",
}

----------------------------------------------------------------- Assassination
local function inExecute()
    return s.HealthPercent() < 35
        and not c.IsTanking()
        and s.TalentRank(14158) == 2 -- Murderous Intent
end

local function doFinisher()
    local cp = GetComboPoints("player")
    return cp == 5 or (not inExecute() and cp == 4)
end

a.spells["Overkill"] = { ID = 58427 }

a.spells["Garrote"] = {
    ID = 703,
}

a.spells["Mutilate"] = {
    ID = 1329,
    CheckFirst = function()
        return GetComboPoints("player") < 4 and not inExecute()
    end
}

a.spells["Backstab during Execute"] = {
    ID = 53,
    CheckFirst = function()
        return GetComboPoints("player") < 5 and inExecute()
    end
}

a.spells["Envenom"] = {
    ID = 32645,
}

a.spells["Smart Envenom"] = {
    ID = 32645,
    CheckFirst = function()
        if not doFinisher() then
            return false
        end
        
        if a.InSoloMode() then
            return true
        end
        
        local rupture = a.spells["Rupture"]
        local refresh = s.MyDebuffDuration(rupture.ID)
        local regen = a.GetRegen()
        local untilCap = s.MaxPower("player") - s.Power("player") - regen
        return refresh > untilCap / regen
            and (not c.HasBuff("Envenom")
                or untilCap <= regen * c.GetBusyTime())
    end
}

a.spells["Rupture for Assassination"] = {
    ID = 1943,
    MyDebuff = 1943,
    EarlyRefresh = 1.9,
    Continue = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        if a.InSoloMode() or not doFinisher() then
            return false
        end
        
        local regen = a.GetRegen()
        local t = c.GetBusyTime()
        local e = s.Power("player") + t * regen
        
        -- sim casting rupture
        t = t + 1
        e = e - 25 + regen
        if GetComboPoints("player") == 5
            and s.TalentRank(14179) == 3 then -- Relentless Strikes
            
            e = e + 25
        end
        
        -- sim generating at least one combo point
        t = t + 1
        if inExecute() then
            e = e - 30 + regen
        else
            e = e - 55 + regen
        end
        
        -- ensure we can cast envenom before SnD wears off
        t = s.BuffDuration(c.GetID("Slice and Dice"), "player") - t
        e = e + t * regen
        return t > .5 and e >= 35
    end
}

a.spells["Vendetta"] = {
    ID = 79140,
    FlashColor = "yellow",
}

a.spells["Vanish for Assassination"] = {
    ID = 1856,
    NoGCD = true,
    FlashColor = "yellow",
    CheckFirst = function()
        return not a.InSoloMode()
            and not c.HasBuff("Overkill", true)
            and not c.HasMyDebuff("Garrote", true)
            and GetComboPoints("player") < 5
    end
}

------------------------------------------------------------------------ Combat
local function sndIsNext()
    -- Can we wait for SnD to wear off without capping?
    -- Only called at 4 or 5 combo points, so no need to consider double combo
    -- point generation.
    local SnD = a.spells["Slice and Dice"]
    local timeTillRefresh = s.BuffDuration(SnD.ID, "player")
    return not a.InSoloMode()
        and s.Power("player")
                 + a.GetRegen() * (timeTillRefresh + 1)
                 - (5 - GetComboPoints("player")) * 39
            < s.MaxPower("player")
end

a.spells["Blade Flurry"] = { ID = 13877 }

a.spells["Adrenaline Rush"] = {
    ID = 13750,
    NoGCD = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        return s.Power("player") < 20
    end
}

a.spells["Killing Spree"] = {
    ID = 51690,
    FlashColor = "yellow",
    NoRangeCheck = 1,
    CheckFirst = function()
        return s.Power("player") < 20
    end
}

a.spells["Expose Armor"] = {
    ID = 8647,
    Debuff = c.ARMOR_DEBUFFS,
    EarlyRefresh = 5,
    Continue = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        return GetComboPoints("player") == 5
            and not sndIsNext()
            and not a.InSoloMode()
    end
}

a.spells["Rupture unless SnD"] = {
    ID = 1943,
    MyDebuff = 1943,
    EarlyRefresh = 1.9,
    RequireDebuff = c.BLEED_DEBUFFS,
    Continue = 1, 
    FlashColor = "yellow",
    CheckFirst = function()
        return GetComboPoints("player") == 5
            and not sndIsNext()
            and not s.Buff("Blade Flurry", "player")
            and not a.InSoloMode()
    end
}

a.spells["Eviscerate for Combat"] = {
    ID = 2098,
    CheckFirst = function()
        return GetComboPoints("player") == 5
            and (not sndIsNext() or a.InSoloMode())
    end
}

a.spells["Revealing Strike"] = {
    ID = 84617,
    MyDebuff = 84617,
    EarlyRefresh = 2,
    CheckFirst = function()
        return GetComboPoints("player") == 4 and not sndIsNext()
    end
}

-- improvement: wait before putting on that 5th combo point to get a better
-- estimate on sndIsNext()
a.spells["Sinister Strike Smart"] = {
    ID = 1752,
    CheckFirst = function()
        local cp = GetComboPoints("player") 
        return cp < 4 or (cp == 4 and sndIsNext())
    end
}


---------------------------------------------------------------------- Subtlety
a.spells["Master of Subtlety"] = { ID = 31665 }
a.spells["Rupture"] = { ID = 1943 }
a.spells["Find Weakness"] = { ID = 91021 }
a.spells["Eviscerate"] = { ID = 2098 }

local function cannotWaitFor(duration)
    duration = duration + 1
    local recupTicks = math.ceil(math.min(
        duration, s.BuffDuration(c.GetID("Recuperate"), "player")) / 3)
    local willPool = a.GetRegen() * duration
        + 4 * recupTicks * s.TalentRank(79152) -- Energetic Recovery
    return willPool >= s.PowerMissing("player")
end

local function cannotWaitForSnD()
    return a.InSoloMode()
        or cannotWaitFor(s.BuffDuration(c.GetID("Slice and Dice"), "player"))
end

local function cannotWaitForRupture()
    return a.InSoloMode()
        or cannotWaitFor(s.MyDebuffDuration(c.GetID("Rupture") - 1.9))
end

local function cannotWaitForRecuperate()
    return cannotWaitFor(s.BuffDuration(c.GetID("Recuperate"), "player") - 2.9)
end

a.spells["Shadow Dance"] = {
    ID = 51713,
    NoGCD = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        return not c.HasBuff("Master of Subtlety", true)
--            and (cannotWaitForRecuperate() or not cannotWaitForRupture())
    end
}

a.spells["Vanish for Subtlety"] = {
    ID = 1856,
    NoGCD = true,
    FlashColor = "yellow",
    CheckFirst = function()
        return not a.InSoloMode()
            and not c.HasBuff("Master of Subtlety", true)
    end
}

a.spells["Preparation"] = {
    ID = 14185,
    FlashColor = "yellow",
    Continue = 1,
    FlashSize = s.FlashSizePercent() / 2,
    CheckFirst = function()
        return c.GetCooldown("Vanish") > 20
            and c.GetCooldown("Shadowstep") > 5
    end
}

a.spells["Premeditation"] = {
    ID = 14183,
    NoGCD = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        return GetComboPoints("player") < 4
    end
}

a.spells["Shadowstep"] = {
    ID = 36554,
    NoGCD = 1,
    FlashColor = "yellow",
    CheckFirst = function()
        return s.Castable(a.spells["Ambush"])
    end
}

local function canFlashRupture(z)
    if a.InSoloMode() then
        return false
    end
    
    local duration = s.MyDebuffDuration(c.GetID("Rupture"))
    local busy = c.GetBusyTime()
    local canEvisIn = math.max(
        busy, (35 - s.Power("player")) / a.GetRegen())
    if duration > canEvisIn + .1 then
        z.FlashID = c.GetID("Eviscerate")
        z.Continue = nil
        z.FlashColor = nil
    else
        z.FlashID = nil
        z.Continue = 1
        z.FlashColor = "yellow"
    end
    return duration - busy < 1.9
end

a.spells["Rupture for Subtlety"] = {
    ID = 1943,
    Override = function(z)
        return cannotWaitForSnD() and canFlashRupture(z)
    end
}

a.spells["Rupture with Master of Subtlety"] = {
    ID = 1943,
    Override = function(z)
        return c.HasBuff("Master of Subtlety") and canFlashRupture(z)
    end
}

a.spells["Recuperate for Energetic Recovery"] = {
    ID = 73651,
    Buff = 73651,
    BuffUnit = "player",
    EarlyRefresh = 2.9,
    NoRangeCheck = 1,
    CheckFirst = function()
        return s.HasTalent(79152) -- Energetic Recovery
            and cannotWaitForSnD()
            and cannotWaitForRupture()
    end
}

a.spells["Eviscerate for Subtlety"] = {
    ID = 2098,
    CheckFirst = function()
        return cannotWaitForSnD()
            and cannotWaitForRupture()
            and cannotWaitForRecuperate()
    end
}

a.spells["Ambush"] = {
    ID = 8676,
}

a.spells["Hemorrhage"] = {
    ID = 16511,
    CheckFirst = function()
        if a.InSoloMode() or c.IsTanking() then
            return true
        end
        
        if not s.HasGlyph(56807) -- Glyph of Hemorrhage
            or c.IsAuraPendingFor("Hemorrhage") then
            
            return false
        end
        
        for i = 1, 500000 do
            local _, _, _, _, _, _, expires, source, _, _, id
                = UnitDebuff("target", i)
            if id == nil then
                return true
            elseif id == 89775 and UnitIsUnit(source, "player") then
                return expires - GetTime() - c.GetBusyTime() < 2.9
            elseif id == 16511 and UnitIsUnit(source, "player") then
                local duration = expires - GetTime()
                if duration > 57 then
                    return false
                end
            end
        end
    end
}

a.spells["Backstab"] = {
    ID = 53,
    CheckFirst = function()
        return not c.IsTanking()
    end
}
